DEFAULT_SEARCH_BACKEND = 'mayan.apps.dynamic_search.backends.django.DjangoSearchBackend'
DEFAULT_SEARCH_BACKEND_ARGUMENTS = {}
DEFAULT_SEARCH_DISABLE_SIMPLE_SEARCH = False
DEFAULT_SEARCH_RESULTS_LIMIT = 100

# Duplicated to keep API compatible until version 4.0
# Merge these two literals and mixins on version 4.0
SEARCH_MODEL_NAME_API_KWARG = 'search_model'
SEARCH_MODEL_NAME_KWARG = 'search_model_name'
TASK_RETRY_DELAY = 5



import uuid

from pytz import common_timezones

from django.conf import settings
from django.db import models
from django.utils.encoding import force_text
from django.utils.translation import ugettext_lazy as _

from .managers import UserLocaleProfileManager


# Required by migration 0010_auto_20180403_0702.py
def upload_to(instance, filename):
    return 'shared-file-{}'.format(uuid.uuid4().hex)


class UserLocaleProfile(models.Model):
    """
    Stores the locale preferences of a user. Stores timezone and language
    at the moment.
    """
    user = models.OneToOneField(
        on_delete=models.CASCADE, related_name='locale_profile',
        to=settings.AUTH_USER_MODEL, verbose_name=_('User')
    )
    timezone = models.CharField(
        choices=zip(common_timezones, common_timezones), max_length=48,
        verbose_name=_('Timezone')
    )
    language = models.CharField(
        choices=settings.LANGUAGES, max_length=8, verbose_name=_('Language')
    )

    objects = UserLocaleProfileManager()

    class Meta:
        verbose_name = _('User locale profile')
        verbose_name_plural = _('User locale profiles')

    def __str__(self):
        return force_text(s=self.user)

    def natural_key(self):
        return self.user.natural_key()
    natural_key.dependencies = [settings.AUTH_USER_MODEL]



import logging

from django.conf import settings
from django.contrib import messages
from django.core.exceptions import ImproperlyConfigured, ValidationError
from django.db import transaction
from django.http import HttpResponseRedirect
from django.urls import reverse
from django.utils.encoding import force_text
from django.utils.translation import ugettext_lazy as _
from django.views.generic import (
    FormView as DjangoFormView, DetailView, TemplateView
)
from django.views.generic.base import View
from django.views.generic.detail import SingleObjectMixin
from django.views.generic.edit import (
    CreateView, DeleteView, FormMixin, ModelFormMixin, UpdateView
)
from django.views.generic.list import ListView

from pure_pagination.mixins import PaginationMixin

from mayan.apps.acls.models import AccessControlList

from .forms import ChoiceForm
from .icons import (
    icon_add_all, icon_confirm_form_cancel, icon_confirm_form_submit,
    icon_remove_all, icon_assign_remove_add,
    icon_assign_remove_remove, icon_sort_down, icon_sort_up
)
from .literals import (
    TEXT_SORT_FIELD_PARAMETER, TEXT_SORT_FIELD_VARIABLE_NAME,
    TEXT_SORT_ORDER_CHOICE_ASCENDING, TEXT_SORT_ORDER_PARAMETER,
    TEXT_SORT_ORDER_VARIABLE_NAME
)
from .mixins import (
    DeleteExtraDataMixin, DownloadMixin, DynamicFormViewMixin,
    ExternalObjectMixin, ExtraContextMixin, FormExtraKwargsMixin,
    ListModeMixin, MultipleObjectMixin, ObjectActionMixin, ObjectNameMixin,
    RedirectionMixin, RestrictedQuerysetMixin, ViewPermissionCheckMixin
)

from .settings import setting_paginate_by

logger = logging.getLogger(name=__name__)


# Required by other views, moved to the top
class MultiFormView(DjangoFormView):
    prefix = None
    prefixes = {}
    template_name = 'appearance/generic_form.html'

    def _create_form(self, form_name, klass):
        form_kwargs = self.get_form_kwargs(form_name=form_name)
        form_create_method = 'create_{}_form'.format(form_name)
        if hasattr(self, form_create_method):
            form = getattr(self, form_create_method)(**form_kwargs)
        else:
            form = klass(**form_kwargs)
        return form

    def all_forms_valid(self, forms):
        return None

    def dispatch(self, request, *args, **kwargs):
        form_classes = self.get_form_classes()
        self.forms = self.get_forms(form_classes=form_classes)
        return super(MultiFormView, self).dispatch(request, *args, **kwargs)

    def forms_invalid(self, forms):
        return self.render_to_response(self.get_context_data(forms=forms))

    def forms_valid(self, forms):
        for form_name, form in forms.items():
            form_valid_method = '{}_form_valid'.format(form_name)

            if hasattr(self, form_valid_method):
                return getattr(self, form_valid_method)(form=form)

        self.all_forms_valid(forms=forms)

        return HttpResponseRedirect(redirect_to=self.get_success_url())

    def get_context_data(self, **kwargs):
        """
        Insert the form into the context dict.
        """
        if 'forms' not in kwargs:
            kwargs['forms'] = self.get_forms(
                form_classes=self.get_form_classes()
            )
        return super(FormMixin, self).get_context_data(**kwargs)

    def get_form_classes(self):
        return self.form_classes

    def get_form_extra_kwargs(self, form_name):
        return None

    def get_form_kwargs(self, form_name):
        kwargs = {}
        kwargs.update({'initial': self.get_initial(form_name=form_name)})
        kwargs.update({'prefix': self.get_prefix(form_name=form_name)})

        if self.request.method in ('POST', 'PUT'):
            kwargs.update(
                {
                    'data': self.request.POST,
                    'files': self.request.FILES,
                }
            )

        kwargs.update(self.get_form_extra_kwargs(form_name=form_name) or {})

        return kwargs

    def get_forms(self, form_classes):
        return dict(
            [
                (
                    key, self._create_form(form_name=key, klass=klass)
                ) for key, klass in form_classes.items()
            ]
        )

    def get_initial(self, form_name):
        initial_method = 'get_{}_initial'.format(form_name)
        if hasattr(self, initial_method):
            return getattr(self, initial_method)()
        else:
            return self.initial.copy()

    def get_prefix(self, form_name):
        return self.prefixes.get(form_name, self.prefix)

    def post(self, request, *args, **kwargs):
        if all([form.is_valid() for form in self.forms.values()]):
            return self.forms_valid(forms=self.forms)
        else:
            return self.forms_invalid(forms=self.forms)


class AddRemoveView(
    ExternalObjectMixin, ExtraContextMixin, ViewPermissionCheckMixin,
    RestrictedQuerysetMixin, MultiFormView
):
    form_classes = {'form_available': ChoiceForm, 'form_added': ChoiceForm}
    list_added_help_text = _(
        'Select entries to be removed. Hold Control to select multiple '
        'entries. Once the selection is complete, click the button below '
        'or double click the list to activate the action.'
    )
    list_available_help_text = _(
        'Select entries to be added. Hold Control to select multiple '
        'entries. Once the selection is complete, click the button below '
        'or double click the list to activate the action.'
    )

    # Form titles
    list_added_title = None
    list_available_title = None

    # Attributes to filter the object to which selections will be added or
    # remove
    main_object_model = None
    main_object_permission = None
    main_object_pk_url_kwarg = None
    main_object_pk_url_kwargs = None
    main_object_source_queryset = None

    # Attributes to filter the queryset of the selection
    secondary_object_model = None
    secondary_object_permission = None
    secondary_object_source_queryset = None

    # Main object methods to use to add and remove selections
    main_object_method_add = None
    main_object_method_remove = None

    # If a method is not specified, use this related field to add and remove
    # selections
    related_field = None

    prefixes = {'form_available': 'available', 'form_added': 'added'}

    def __init__(self, *args, **kwargs):
        self.external_object_class = self.main_object_model
        self.external_object_permission = self.main_object_permission
        self.external_object_pk_url_kwarg = self.main_object_pk_url_kwarg
        self.external_object_pk_url_kwargs = self.main_object_pk_url_kwargs
        self.external_object_queryset = self.main_object_source_queryset

        super(AddRemoveView, self).__init__(*args, **kwargs)

    def _action_add(self, queryset):
        kwargs = {'queryset': queryset}
        kwargs.update(self.get_action_add_extra_kwargs())
        kwargs.update(self.get_actions_extra_kwargs())

        if hasattr(self, 'action_add'):
            with transaction.atomic():
                self.action_add(**kwargs)
        elif self.main_object_method_add:
            getattr(self.main_object, self.main_object_method_add)(**kwargs)
        elif self.related_field:
            getattr(self.main_object, self.related_field).add(*queryset)
        else:
            raise ImproperlyConfigured(
                'View {} must be called with a main_object_method_add, a '
                'related_field, or an action_add '
                'method.'.format(self.__class__.__name__)
            )

    def _action_remove(self, queryset):
        kwargs = {'queryset': queryset}
        kwargs.update(self.get_action_remove_extra_kwargs())
        kwargs.update(self.get_actions_extra_kwargs())

        if hasattr(self, 'action_remove'):
            with transaction.atomic():
                self.action_remove(**kwargs)
        elif self.main_object_method_remove:
            getattr(self.main_object, self.main_object_method_remove)(**kwargs)
        elif self.related_field:
            getattr(self.main_object, self.related_field).remove(*queryset)
        else:
            raise ImproperlyConfigured(
                'View {} must be called with a main_object_method_remove, a '
                'related_field, or an action_remove '
                'method.'.format(self.__class__.__name__)
            )

    def dispatch(self, request, *args, **kwargs):
        self.main_object = self.get_external_object()
        result = super(AddRemoveView, self).dispatch(request=request, *args, **kwargs)
        return result

    def forms_valid(self, forms):
        if 'available-add_all' in self.request.POST:
            selection_add = self.get_list_available_queryset()
        else:
            selection_add = self.get_list_available_queryset().filter(
                pk__in=forms['form_available'].cleaned_data['selection']
            )

        self._action_add(queryset=selection_add)

        if 'added-remove_all' in self.request.POST:
            selection_remove = self.get_list_added_queryset()
        else:
            selection_remove = self.get_list_added_queryset().filter(
                pk__in=forms['form_added'].cleaned_data['selection']
            )

        self._action_remove(queryset=selection_remove)

        return super(AddRemoveView, self).forms_valid(forms=forms)

    def generate_choices(self, queryset):
        for obj in queryset:
            yield (obj.pk, force_text(s=obj))

    def get_action_add_extra_kwargs(self):
        # Keyword arguments to apply to the add method
        return {}

    def get_action_remove_extra_kwargs(self):
        # Keyword arguments to apply to the remove method
        return {}

    def get_actions_extra_kwargs(self):
        # Keyword arguments to apply to both the add and remove methods
        return {}

    def get_context_data(self, **kwargs):
        # Use get_context_data to leave the get_extra_context for subclasses
        context = super(AddRemoveView, self).get_context_data(**kwargs)
        context.update(
            {
                'subtemplates_list': [
                    {
                        'name': 'appearance/generic_form_subtemplate.html',
                        'column_class': 'col-xs-12 col-sm-6 col-md-6 col-lg-6',
                        'context': {
                            'extra_buttons': [
                                {
                                    'label': _('Add all'),
                                    'icon_class': icon_add_all,
                                    'name': 'add_all',
                                }
                            ],
                            'form': self.forms['form_available'],
                            'form_css_classes': 'form-hotkey-double-click',
                            'hide_labels': True,
                            'submit_icon_class': icon_assign_remove_add,
                            'submit_label': _('Add'),
                            'title': self.list_available_title or ' ',
                        }
                    },
                    {
                        'name': 'appearance/generic_form_subtemplate.html',
                        'column_class': 'col-xs-12 col-sm-6 col-md-6 col-lg-6',
                        'context': {
                            'extra_buttons': [
                                {
                                    'label': _('Remove all'),
                                    'icon_class': icon_remove_all,
                                    'name': 'remove_all',
                                }
                            ],
                            'form': self.forms['form_added'],
                            'form_css_classes': 'form-hotkey-double-click',
                            'hide_labels': True,
                            'submit_icon_class': icon_assign_remove_remove,
                            'submit_label': _('Remove'),
                            'title': self.list_added_title or ' ',
                        }
                    }
                ]
            }
        )

        return context

    def get_disabled_choices(self):
        return ()

    def get_form_extra_kwargs(self, form_name):
        if form_name == 'form_available':
            return {
                'choices': self.generate_choices(
                    queryset=self.get_list_available_queryset()
                ),
                'help_text': self.get_list_available_help_text()
            }
        else:
            return {
                'choices': self.generate_choices(
                    queryset=self.get_list_added_queryset()
                ),
                'disabled_choices': self.get_disabled_choices(),
                'help_text': self.get_list_added_help_text()
            }

    def get_list_added_help_text(self):
        return self.list_added_help_text

    def get_list_added_queryset(self):
        if not self.related_field:
            raise ImproperlyConfigured(
                'View {} must be called with either a related_field or '
                'override .get_list_added_queryset().'.format(
                    self.__class__.__name__
                )
            )

        return self.get_secondary_object_list().filter(
            pk__in=getattr(self.main_object, self.related_field).values('pk')
        )

    def get_list_available_help_text(self):
        return self.list_available_help_text

    def get_list_available_queryset(self):
        return self.get_secondary_object_list().exclude(
            pk__in=self.get_list_added_queryset().values('pk')
        )

    def get_secondary_object_list(self):
        queryset = self.get_secondary_object_source_queryset()

        if queryset is None:
            queryset = self.secondary_object_model._meta.default_manager.all()

        if self.secondary_object_permission:
            return AccessControlList.objects.restrict_queryset(
                permission=self.secondary_object_permission, queryset=queryset,
                user=self.request.user
            )
        else:
            return queryset

    def get_secondary_object_source_queryset(self):
        return self.secondary_object_source_queryset

    def get_success_url(self):
        # Redirect to the same view
        return reverse(
            viewname=self.request.resolver_match.view_name,
            kwargs=self.request.resolver_match.kwargs
        )


class ConfirmView(
    RestrictedQuerysetMixin, ViewPermissionCheckMixin, ExtraContextMixin,
    RedirectionMixin, TemplateView
):
    template_name = 'appearance/generic_confirm.html'

    def get_context_data(self, **kwargs):
        context = {
            'submit_icon_class': icon_confirm_form_submit,
            'cancel_icon_class': icon_confirm_form_cancel
        }
        context.update(super(ConfirmView, self).get_context_data(**kwargs))
        return context

    def post(self, request, *args, **kwargs):
        self.view_action()
        return HttpResponseRedirect(redirect_to=self.get_success_url())


class FormView(
    ViewPermissionCheckMixin, ExtraContextMixin, RedirectionMixin,
    FormExtraKwargsMixin, DjangoFormView
):
    template_name = 'appearance/generic_form.html'


class DynamicFormView(DynamicFormViewMixin, FormView):
    pass


class MultipleObjectFormActionView(
    ExtraContextMixin, ObjectActionMixin, ViewPermissionCheckMixin,
    RestrictedQuerysetMixin, MultipleObjectMixin, FormExtraKwargsMixin,
    RedirectionMixin, DjangoFormView
):
    """
    This view will present a form and upon receiving a POST request will
    perform an action on an object or queryset
    """
    template_name = 'appearance/generic_form.html'

    def __init__(self, *args, **kwargs):
        result = super(MultipleObjectFormActionView, self).__init__(*args, **kwargs)

        if self.__class__.mro()[0].get_queryset != MultipleObjectFormActionView.get_queryset:
            raise ImproperlyConfigured(
                '%(cls)s is overloading the get_queryset method. Subclasses '
                'should implement the get_source_queryset method instead. ' % {
                    'cls': self.__class__.__name__
                }
            )

        return result

    def form_invalid(self, form):
        if settings.DEBUG or settings.TESTING:
            logger.debug('Form invalid; %s', form.errors)

        return super().form_invalid(form=form)

    def form_valid(self, form):
        self.view_action(form=form)
        return super(MultipleObjectFormActionView, self).form_valid(form=form)

    def get_queryset(self):
        try:
            return super(MultipleObjectFormActionView, self).get_queryset()
        except ImproperlyConfigured:
            self.queryset = self.get_source_queryset()
            return super(MultipleObjectFormActionView, self).get_queryset()


class MultipleObjectConfirmActionView(
    ExtraContextMixin, ObjectActionMixin, ViewPermissionCheckMixin,
    RestrictedQuerysetMixin, MultipleObjectMixin, RedirectionMixin, TemplateView
):
    template_name = 'appearance/generic_confirm.html'

    def __init__(self, *args, **kwargs):
        result = super(MultipleObjectConfirmActionView, self).__init__(*args, **kwargs)

        if self.__class__.mro()[0].get_queryset != MultipleObjectConfirmActionView.get_queryset:
            raise ImproperlyConfigured(
                '%(cls)s is overloading the get_queryset method. Subclasses '
                'should implement the get_source_queryset method instead. ' % {
                    'cls': self.__class__.__name__
                }
            )

        return result

    def get_queryset(self):
        try:
            return super(MultipleObjectConfirmActionView, self).get_queryset()
        except ImproperlyConfigured:
            self.queryset = self.get_source_queryset()
            return super(MultipleObjectConfirmActionView, self).get_queryset()

    def post(self, request, *args, **kwargs):
        self.view_action()
        return HttpResponseRedirect(redirect_to=self.get_success_url())


class SimpleView(ViewPermissionCheckMixin, ExtraContextMixin, TemplateView):
    """
    Basic template view class with permission check and extra context
    """


class SingleObjectCreateView(
    ObjectNameMixin, ViewPermissionCheckMixin, ExtraContextMixin,
    RedirectionMixin, FormExtraKwargsMixin, CreateView
):
    error_message_duplicate = None
    template_name = 'appearance/generic_form.html'

    def form_valid(self, form):
        # This overrides the original Django form_valid method

        self.object = form.save(commit=False)

        if hasattr(self, 'get_instance_extra_data'):
            for key, value in self.get_instance_extra_data().items():
                setattr(self.object, key, value)

        if hasattr(self, 'get_save_extra_data'):
            save_extra_data = self.get_save_extra_data()
        else:
            save_extra_data = {}

        # Validate duplicates first
        try:
            self.object.validate_unique()
        except ValidationError as exception:
            context = self.get_context_data()

            error_message = self.get_error_message_duplicate() or _(
                'Duplicate data error: %(error)s'
            ) % {
                'error': '\n'.join(exception.messages)
            }

            messages.error(
                message=error_message, request=self.request
            )
            return super(
                SingleObjectCreateView, self
            ).form_invalid(form=form)

        try:
            self.object.save(**save_extra_data)
        except Exception as exception:
            if settings.DEBUG or settings.TESTING:
                raise
            else:
                context = self.get_context_data()

                messages.error(
                    message=_('%(object)s not created, error: %(error)s') % {
                        'object': self.get_object_name(context=context),
                        'error': exception
                    }, request=self.request
                )
                return super(
                    SingleObjectCreateView, self
                ).form_invalid(form=form)
        else:
            context = self.get_context_data()

            messages.success(
                message=_(
                    '%(object)s created successfully.'
                ) % {'object': self.get_object_name(context=context)},
                request=self.request
            )

        return HttpResponseRedirect(redirect_to=self.get_success_url())

    def get_error_message_duplicate(self):
        return self.error_message_duplicate


class SingleObjectDeleteView(
    ObjectNameMixin, DeleteExtraDataMixin, ViewPermissionCheckMixin,
    RestrictedQuerysetMixin, ExtraContextMixin, RedirectionMixin, DeleteView
):
    template_name = 'appearance/generic_confirm.html'

    def __init__(self, *args, **kwargs):
        result = super(SingleObjectDeleteView, self).__init__(*args, **kwargs)

        if self.__class__.mro()[0].get_queryset != SingleObjectDeleteView.get_queryset:
            raise ImproperlyConfigured(
                '%(cls)s is overloading the get_queryset method. Subclasses '
                'should implement the get_source_queryset method instead. ' % {
                    'cls': self.__class__.__name__
                }
            )

        return result

    def delete(self, request, *args, **kwargs):
        self.object = self.get_object()
        context = self.get_context_data()
        object_name = self.get_object_name(context=context)

        try:
            result = super(SingleObjectDeleteView, self).delete(request, *args, **kwargs)
        except Exception as exception:
            messages.error(
                message=_('%(object)s not deleted, error: %(error)s.') % {
                    'object': object_name,
                    'error': exception
                }, request=self.request
            )
            raise
        else:
            messages.success(
                message=_(
                    '%(object)s deleted successfully.'
                ) % {'object': object_name},
                request=self.request
            )

            return result

    def get_context_data(self, **kwargs):
        context = super(SingleObjectDeleteView, self).get_context_data(**kwargs)
        context.update({'delete_view': True})
        return context

    def get_queryset(self):
        try:
            return super(SingleObjectDeleteView, self).get_queryset()
        except ImproperlyConfigured:
            self.queryset = self.get_source_queryset()
            return super(SingleObjectDeleteView, self).get_queryset()


class SingleObjectDetailView(
    ViewPermissionCheckMixin, RestrictedQuerysetMixin, FormExtraKwargsMixin,
    ExtraContextMixin, ModelFormMixin, DetailView
):
    template_name = 'appearance/generic_form.html'

    def __init__(self, *args, **kwargs):
        result = super(SingleObjectDetailView, self).__init__(*args, **kwargs)

        if self.__class__.mro()[0].get_queryset != SingleObjectDetailView.get_queryset:
            raise ImproperlyConfigured(
                '%(cls)s is overloading the get_queryset method. Subclasses '
                'should implement the get_source_queryset method instead. ' % {
                    'cls': self.__class__.__name__
                }
            )

        return result

    def get_context_data(self, **kwargs):
        context = super(SingleObjectDetailView, self).get_context_data(**kwargs)
        context.update({'read_only': True, 'form': self.get_form()})
        return context

    def get_queryset(self):
        try:
            return super(SingleObjectDetailView, self).get_queryset()
        except ImproperlyConfigured:
            self.queryset = self.get_source_queryset()
            return super(SingleObjectDetailView, self).get_queryset()


class BaseDownloadView(DownloadMixin, ViewPermissionCheckMixin, View):
    def get(self, request, *args, **kwargs):
        return self.render_to_response()


class SingleObjectDownloadView(
    RestrictedQuerysetMixin, SingleObjectMixin, BaseDownloadView
):
    def get(self, request, *args, **kwargs):
        self.object = self.get_object()
        return super(SingleObjectDownloadView, self).get(
            request, *args, **kwargs
        )

    def get_download_file_object(self):
        return self.object.open()

    def get_download_label(self):
        return force_text(s=self.object)


class MultipleObjectDownloadView(
    RestrictedQuerysetMixin, MultipleObjectMixin, BaseDownloadView
):
    """
    View that support receiving multiple objects via a pk_list query.
    """
    def __init__(self, *args, **kwargs):
        result = super(MultipleObjectDownloadView, self).__init__(*args, **kwargs)

        if self.__class__.mro()[0].get_queryset != MultipleObjectDownloadView.get_queryset:
            raise ImproperlyConfigured(
                '%(cls)s is overloading the get_queryset method. Subclasses '
                'should implement the get_source_queryset method instead. ' % {
                    'cls': self.__class__.__name__
                }
            )

        return result

    def get_queryset(self):
        try:
            return super(MultipleObjectDownloadView, self).get_queryset()
        except ImproperlyConfigured:
            self.queryset = self.get_source_queryset()
            return super(MultipleObjectDownloadView, self).get_queryset()


class SingleObjectDynamicFormCreateView(
    DynamicFormViewMixin, SingleObjectCreateView
):
    pass


class SingleObjectEditView(
    ObjectNameMixin, ViewPermissionCheckMixin, RestrictedQuerysetMixin,
    ExtraContextMixin, FormExtraKwargsMixin, RedirectionMixin, UpdateView
):
    template_name = 'appearance/generic_form.html'

    def form_valid(self, form):
        # This overrides the original Django form_valid method

        self.object = form.save(commit=False)

        if hasattr(self, 'get_instance_extra_data'):
            for key, value in self.get_instance_extra_data().items():
                setattr(self.object, key, value)

        if hasattr(self, 'get_save_extra_data'):
            save_extra_data = self.get_save_extra_data()
        else:
            save_extra_data = {}

        context = self.get_context_data()
        object_name = self.get_object_name(context=context)

        try:
            self.object.save(**save_extra_data)
        except Exception as exception:
            if settings.DEBUG or settings.TESTING:
                raise
            else:
                messages.error(
                    message=_('%(object)s not updated, error: %(error)s.') % {
                        'object': object_name,
                        'error': exception
                    }, request=self.request
                )
                return super(
                    SingleObjectEditView, self
                ).form_invalid(form=form)
        else:
            messages.success(
                message=_(
                    '%(object)s updated successfully.'
                ) % {'object': object_name}, request=self.request
            )

        return HttpResponseRedirect(redirect_to=self.get_success_url())

    def get_object(self, queryset=None):
        obj = super(SingleObjectEditView, self).get_object(queryset=queryset)

        if hasattr(self, 'get_instance_extra_data'):
            for key, value in self.get_instance_extra_data().items():
                setattr(obj, key, value)

        return obj


class SingleObjectDynamicFormEditView(
    DynamicFormViewMixin, SingleObjectEditView
):
    pass


class SingleObjectListView(
    ListModeMixin, PaginationMixin, ViewPermissionCheckMixin,
    RestrictedQuerysetMixin, ExtraContextMixin, RedirectionMixin, ListView
):
    template_name = 'appearance/generic_list.html'

    def __init__(self, *args, **kwargs):
        result = super(SingleObjectListView, self).__init__(*args, **kwargs)

        if self.__class__.mro()[0].get_queryset != SingleObjectListView.get_queryset:
            raise ImproperlyConfigured(
                '%(cls)s is overloading the get_queryset method. Subclasses '
                'should implement the get_source_queryset method instead. ' % {
                    'cls': self.__class__.__name__
                }
            )

        return result

    def get_context_data(self, **kwargs):
        context = super(SingleObjectListView, self).get_context_data(**kwargs)

        context.update(
            {
                TEXT_SORT_FIELD_VARIABLE_NAME: self.get_sort_field(),
                TEXT_SORT_ORDER_VARIABLE_NAME: self.get_sort_order(),
                'icon_sort': self.get_sort_icon(),
            }
        )
        return context

    def get_paginate_by(self, queryset):
        return setting_paginate_by.value

    def get_queryset(self):
        try:
            queryset = super(SingleObjectListView, self).get_queryset()
        except ImproperlyConfigured:
            self.queryset = self.get_source_queryset()
            queryset = super(SingleObjectListView, self).get_queryset()

        self.field_name = self.get_sort_field()
        if self.get_sort_order() == TEXT_SORT_ORDER_CHOICE_ASCENDING:
            sort_order = ''
        else:
            sort_order = '-'

        if self.field_name:
            queryset = queryset.order_by(
                '{}{}'.format(sort_order, self.field_name)
            )

        return queryset

    def get_sort_field(self):
        return self.request.GET.get(TEXT_SORT_FIELD_PARAMETER)

    def get_sort_icon(self):
        sort_order = self.get_sort_order()
        if not sort_order:
            return
        elif sort_order == TEXT_SORT_ORDER_CHOICE_ASCENDING:
            return icon_sort_down
        else:
            return icon_sort_up

    def get_sort_order(self):
        return self.request.GET.get(TEXT_SORT_ORDER_PARAMETER)



from rest_framework import status

from mayan.apps.documents.permissions import permission_document_type_edit
from mayan.apps.documents.tests.base import DocumentTestMixin
from mayan.apps.rest_api.tests.base import BaseAPITestCase

from ..models import ResolvedWebLink, WebLink
from ..permissions import (
    permission_web_link_create, permission_web_link_delete,
    permission_web_link_edit, permission_web_link_instance_view,
    permission_web_link_view
)

from .literals import TEST_WEB_LINK_LABEL_EDITED, TEST_WEB_LINK_LABEL
from .mixins import (
    ResolvedWebLinkAPIViewTestMixin, WebLinkAPIViewTestMixin, WebLinkTestMixin
)


class WebLinkAPIViewTestCase(
    DocumentTestMixin, WebLinkTestMixin, WebLinkAPIViewTestMixin,
    BaseAPITestCase
):
    auto_create_test_document_type = False
    auto_upload_test_document = False

    def test_web_link_create_view_no_permission(self):
        response = self._request_test_web_link_create_api_view()
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertEqual(WebLink.objects.count(), 0)

    def test_web_link_create_view_with_permission(self):
        self.grant_permission(permission=permission_web_link_create)

        response = self._request_test_web_link_create_api_view()
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        web_link = WebLink.objects.first()
        self.assertEqual(response.data['id'], web_link.pk)
        self.assertEqual(response.data['label'], TEST_WEB_LINK_LABEL)

        self.assertEqual(WebLink.objects.count(), 1)
        self.assertEqual(web_link.label, TEST_WEB_LINK_LABEL)

    def test_web_link_create_with_document_types_view_no_permission(self):
        self._create_test_document_type()

        response = self._request_test_web_link_create_with_document_type_api_view()
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertEqual(WebLink.objects.count(), 0)

    def test_web_link_create_with_document_types_view_web_link_permission(self):
        self._create_test_document_type()
        self.grant_permission(permission=permission_web_link_create)

        response = self._request_test_web_link_create_with_document_type_api_view()
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        web_link = WebLink.objects.first()
        self.assertEqual(response.data['id'], web_link.pk)
        self.assertEqual(response.data['label'], TEST_WEB_LINK_LABEL)

        self.assertEqual(WebLink.objects.count(), 1)
        self.assertEqual(web_link.label, TEST_WEB_LINK_LABEL)
        self.assertTrue(
            self.test_document_type not in web_link.document_types.all()
        )

    def test_web_link_create_with_document_types_view_with_document_type_access(self):
        self._create_test_document_type()
        self.grant_access(
            obj=self.test_document_type, permission=permission_document_type_edit
        )
        response = self._request_test_web_link_create_with_document_type_api_view()
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertEqual(WebLink.objects.count(), 0)

    def test_web_link_create_with_document_types_view_with_full_access(self):
        self._create_test_document_type()
        self.grant_permission(permission=permission_web_link_create)
        self.grant_access(
            obj=self.test_document_type, permission=permission_document_type_edit
        )

        response = self._request_test_web_link_create_with_document_type_api_view()
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        web_link = WebLink.objects.first()
        self.assertEqual(response.data['id'], web_link.pk)
        self.assertEqual(response.data['label'], TEST_WEB_LINK_LABEL)

        self.assertEqual(WebLink.objects.count(), 1)
        self.assertEqual(web_link.label, TEST_WEB_LINK_LABEL)
        self.assertTrue(
            self.test_document_type in web_link.document_types.all()
        )

    def test_web_link_delete_view_no_permission(self):
        self._create_test_web_link()

        response = self._request_test_web_link_delete_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(WebLink.objects.count(), 1)

    def test_web_link_delete_view_with_access(self):
        self._create_test_web_link()
        self.grant_access(
            obj=self.test_web_link, permission=permission_web_link_delete
        )

        response = self._request_test_web_link_delete_api_view()
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        self.assertEqual(WebLink.objects.count(), 0)

    def test_web_link_detail_view_no_permission(self):
        self._create_test_web_link()

        response = self._request_test_web_link_detail_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertFalse('label' in response.data)

    def test_web_link_detail_view_with_access(self):
        self._create_test_web_link()
        self.grant_access(
            obj=self.test_web_link, permission=permission_web_link_view
        )

        response = self._request_test_web_link_detail_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertEqual(
            response.data['label'], TEST_WEB_LINK_LABEL
        )

    def test_web_link_edit_view_via_patch_no_permission(self):
        self._create_test_document_type()
        self._create_test_web_link()

        response = self._request_test_web_link_edit_patch_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.test_web_link.refresh_from_db()
        self.assertEqual(self.test_web_link.label, TEST_WEB_LINK_LABEL)

    def test_web_link_edit_view_via_patch_with_access(self):
        self._create_test_document_type()
        self._create_test_web_link()
        self.grant_access(
            obj=self.test_web_link, permission=permission_web_link_edit
        )

        response = self._request_test_web_link_edit_patch_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.test_web_link.refresh_from_db()
        self.assertEqual(
            self.test_web_link.label, TEST_WEB_LINK_LABEL_EDITED
        )

    def test_web_link_edit_view_via_put_no_permission(self):
        self._create_test_document_type()
        self._create_test_web_link()

        response = self._request_test_web_link_edit_put_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.test_web_link.refresh_from_db()
        self.assertEqual(self.test_web_link.label, TEST_WEB_LINK_LABEL)

    def test_web_link_edit_view_via_put_with_access(self):
        self._create_test_document_type()
        self._create_test_web_link()
        self.grant_access(
            obj=self.test_web_link, permission=permission_web_link_edit
        )

        response = self._request_test_web_link_edit_put_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.test_web_link.refresh_from_db()
        self.assertEqual(
            self.test_web_link.label, TEST_WEB_LINK_LABEL_EDITED
        )


class ResolvedWebLinkAPIViewTestCase(
    DocumentTestMixin, WebLinkTestMixin, ResolvedWebLinkAPIViewTestMixin,
    BaseAPITestCase
):
    def setUp(self):
        super(ResolvedWebLinkAPIViewTestCase, self).setUp()
        self._create_test_web_link(add_document_type=True)

    def test_resolved_web_link_detail_view_no_permission(self):
        response = self._request_resolved_web_link_detail_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_resolved_web_link_detail_view_with_document_access(self):
        self.grant_access(
            obj=self.test_document,
            permission=permission_web_link_instance_view
        )

        response = self._request_resolved_web_link_detail_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_resolved_web_link_detail_view_with_web_link_access(self):
        self.grant_access(
            obj=self.test_web_link,
            permission=permission_web_link_instance_view
        )

        response = self._request_resolved_web_link_detail_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_resolved_web_link_detail_view_with_full_access(self):
        self.grant_access(
            obj=self.test_web_link,
            permission=permission_web_link_instance_view
        )
        self.grant_access(
            obj=self.test_document,
            permission=permission_web_link_instance_view
        )

        response = self._request_resolved_web_link_detail_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(
            response.data['id'], self.test_web_link.pk
        )

    def test_resolved_web_link_list_view_no_permission(self):
        response = self._request_resolved_web_link_list_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertFalse('results' in response.data)

    def test_resolved_web_link_list_view_with_document_access(self):
        self.grant_access(
            obj=self.test_document,
            permission=permission_web_link_instance_view
        )

        response = self._request_resolved_web_link_list_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertEqual(response.data['count'], 0)

    def test_resolved_web_link_list_view_with_web_link_access(self):
        self.grant_access(
            obj=self.test_web_link,
            permission=permission_web_link_instance_view
        )
        response = self._request_resolved_web_link_list_view()

        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertFalse('results' in response.data)

    def test_resolved_web_link_list_view_with_full_access(self):
        self.grant_access(
            obj=self.test_web_link,
            permission=permission_web_link_instance_view
        )
        self.grant_access(
            obj=self.test_document,
            permission=permission_web_link_instance_view
        )

        response = self._request_resolved_web_link_list_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(
            response.data['results'][0]['id'], self.test_web_link.pk
        )

    def test_resolved_web_link_navigate_view_no_permission(self):
        response = self._request_resolved_web_link_navigate_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_resolved_web_link_navigate_view_with_web_link_access(self):
        self.grant_access(
            obj=self.test_web_link,
            permission=permission_web_link_instance_view
        )

        response = self._request_resolved_web_link_navigate_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_resolved_web_link_navigate_view_with_document_access(self):
        self.grant_access(
            obj=self.test_document,
            permission=permission_web_link_instance_view
        )
        response = self._request_resolved_web_link_navigate_view()

        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_resolved_web_link_navigate_view_with_full_access(self):
        self.grant_access(
            obj=self.test_document,
            permission=permission_web_link_instance_view
        )
        self.grant_access(
            obj=self.test_web_link,
            permission=permission_web_link_instance_view
        )

        response = self._request_resolved_web_link_navigate_view()
        self.assertEqual(response.status_code, status.HTTP_302_FOUND)
        self.assertEqual(
            response.url, ResolvedWebLink.objects.get(
                pk=self.test_web_link.pk
            ).get_redirect_url_for(document=self.test_document)
        )



from mayan.apps.documents.permissions import permission_document_view
from mayan.apps.documents.tests.base import GenericDocumentViewTestCase

from ..permissions import permission_workflow_view

from .mixins import (
    WorkflowRuntimeProxyStateViewTestMixin,
    WorkflowRuntimeProxyViewTestMixin, WorkflowTestMixin
)


class WorkflowRuntimeProxyViewTestCase(
    WorkflowRuntimeProxyViewTestMixin, WorkflowTestMixin,
    GenericDocumentViewTestCase
):
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()
        self._create_test_workflow(add_document_type=True)
        self._create_test_workflow_states()
        self._create_test_workflow_transition()
        self._create_test_document_stub()

    def test_workflow_runtime_proxy_document_list_view_no_permission(self):
        response = self._request_test_workflow_runtime_proxy_document_list_view()
        self.assertNotContains(
            response=response, text=self.test_workflow.label, status_code=404
        )
        self.assertNotContains(
            response=response, text=self.test_document.label, status_code=404
        )

    def test_workflow_runtime_proxy_document_list_view_with_workflow_access(self):
        self.grant_access(
            obj=self.test_workflow, permission=permission_workflow_view
        )

        response = self._request_test_workflow_runtime_proxy_document_list_view()
        self.assertContains(
            response=response, text=self.test_workflow.label, status_code=200
        )
        self.assertNotContains(
            response=response, text=self.test_document.label, status_code=200
        )

    def test_workflow_runtime_proxy_document_list_view_with_document_access(self):
        self.grant_access(
            obj=self.test_document, permission=permission_document_view
        )

        response = self._request_test_workflow_runtime_proxy_document_list_view()
        self.assertNotContains(
            response=response, text=self.test_workflow.label, status_code=404
        )
        self.assertNotContains(
            response=response, text=self.test_document.label, status_code=404
        )

    def test_workflow_runtime_proxy_document_list_view_with_full_access(self):
        self.grant_access(
            obj=self.test_document, permission=permission_document_view
        )
        self.grant_access(
            obj=self.test_workflow, permission=permission_workflow_view
        )

        response = self._request_test_workflow_runtime_proxy_document_list_view()
        self.assertContains(
            response=response, text=self.test_workflow.label, status_code=200
        )
        self.assertContains(
            response=response, text=self.test_document.label, status_code=200
        )

    def test_trashed_document_workflow_runtime_proxy_document_list_view_with_full_access(self):
        self.grant_access(
            obj=self.test_document, permission=permission_document_view
        )
        self.grant_access(
            obj=self.test_workflow, permission=permission_workflow_view
        )

        self.test_document.delete()

        response = self._request_test_workflow_runtime_proxy_document_list_view()
        self.assertContains(
            response=response, text=self.test_workflow.label, status_code=200
        )
        self.assertNotContains(
            response=response, text=self.test_document.label, status_code=200
        )

    def test_workflow_runtime_proxy_list_view_no_permission(self):
        response = self._request_test_workflow_runtime_proxy_list_view()
        self.assertNotContains(
            response=response, text=self.test_workflow.label, status_code=200
        )

    def test_workflow_runtime_proxy_list_view_with_access(self):
        self.grant_access(
            obj=self.test_workflow, permission=permission_workflow_view
        )

        response = self._request_test_workflow_runtime_proxy_list_view()
        self.assertContains(
            response=response, text=self.test_workflow.label, status_code=200
        )


class WorkflowRuntimeProxyStateViewTestCase(
    WorkflowRuntimeProxyStateViewTestMixin, WorkflowTestMixin,
    GenericDocumentViewTestCase
):
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()
        self._create_test_workflow(add_document_type=True)
        self._create_test_workflow_states()
        self._create_test_workflow_transition()
        self._create_test_document_stub()

    def test_workflow_runtime_proxy_state_document_list_view_no_permission(self):
        response = self._request_test_workflow_runtime_proxy_state_document_list_view()
        self.assertNotContains(
            response=response, text=self.test_workflow_state_1.label,
            status_code=404
        )
        self.assertNotContains(
            response=response, text=self.test_document.label, status_code=404
        )
        self.assertNotContains(
            response=response, text=self.test_document.label, status_code=404
        )

    def test_workflow_runtime_proxy_state_document_list_view_with_workflow_access(self):
        self.grant_access(
            obj=self.test_workflow, permission=permission_workflow_view
        )

        response = self._request_test_workflow_runtime_proxy_state_document_list_view()
        self.assertContains(
            response=response, text=self.test_workflow_state_1.label,
            status_code=200
        )
        self.assertNotContains(
            response=response, text=self.test_document.label, status_code=200
        )

    def test_workflow_runtime_proxy_state_document_list_view_with_document_access(self):
        self.grant_access(
            obj=self.test_document, permission=permission_document_view
        )

        response = self._request_test_workflow_runtime_proxy_state_document_list_view()
        self.assertNotContains(
            response=response, text=self.test_workflow_state_1.label,
            status_code=404
        )
        self.assertNotContains(
            response=response, text=self.test_document.label, status_code=404
        )

    def test_workflow_runtime_proxy_state_document_list_view_with_full_access(self):
        self.grant_access(
            obj=self.test_document, permission=permission_document_view
        )
        self.grant_access(
            obj=self.test_workflow, permission=permission_workflow_view
        )

        response = self._request_test_workflow_runtime_proxy_state_document_list_view()
        self.assertContains(
            response=response, text=self.test_workflow_state_1.label,
            status_code=200
        )
        self.assertContains(
            response=response, text=self.test_document.label, status_code=200
        )

    def test_trashed_document_workflow_runtime_proxy_state_document_list_view_with_access(self):
        self.grant_access(
            obj=self.test_document, permission=permission_document_view
        )
        self.grant_access(
            obj=self.test_workflow, permission=permission_workflow_view
        )

        self.test_document.delete()

        response = self._request_test_workflow_runtime_proxy_state_document_list_view()
        self.assertContains(
            response=response, text=self.test_workflow_state_1.label,
            status_code=200
        )
        self.assertNotContains(
            response=response, text=self.test_document.label, status_code=200
        )

    def test_workflow_runtime_proxy_state_list_view_no_permission(self):
        response = self._request_test_workflow_runtime_proxy_state_list_view()
        self.assertNotContains(
            response=response, text=self.test_workflow_state_1.label,
            status_code=404
        )

    def test_workflow_runtime_proxy_state_list_view_with_access(self):
        self.grant_access(
            obj=self.test_workflow, permission=permission_workflow_view
        )

        response = self._request_test_workflow_runtime_proxy_state_list_view()
        self.assertContains(
            response=response, text=self.test_workflow_state_1.label,
            status_code=200
        )



from django.utils.translation import ugettext_lazy as _

from mayan.apps.smart_settings.classes import SettingNamespace

from .literals import (
    DEFAULT_COMMON_DB_SYNC_TASK_DELAY, DEFAULT_COMMON_DISABLED_APPS,
    DEFAULT_COMMON_EXTRA_APPS, DEFAULT_COMMON_HOME_VIEW,
    DEFAULT_COMMON_PROJECT_TITLE, DEFAULT_COMMON_PROJECT_URL,
    DEFAULT_COMMON_URL_BASE_PATH
)

namespace = SettingNamespace(
    label=_('Common'), name='common', version='0002'
)

settings_db_sync_task_delay = namespace.add_setting(
    default=DEFAULT_COMMON_DB_SYNC_TASK_DELAY,
    global_name='COMMON_DB_SYNC_TASK_DELAY', help_text=_(
        'Time to delay background tasks that depend on a database commit to '
        'propagate.'
    )
)
setting_disabled_apps = namespace.add_setting(
    default=DEFAULT_COMMON_DISABLED_APPS,
    global_name='COMMON_DISABLED_APPS', help_text=_(
        'A list of strings designating all applications that are to be '
        'removed from the list normally installed by Mayan EDMS. Each '
        'string should be a dotted Python path to: an application '
        'configuration class (preferred), or a package containing an '
        'application.'
    )
)
setting_extra_apps = namespace.add_setting(
    default=DEFAULT_COMMON_EXTRA_APPS, global_name='COMMON_EXTRA_APPS',
    help_text=_(
        'A list of strings designating all applications that are installed '
        'beyond those normally installed by Mayan EDMS. Each string '
        'should be a dotted Python path to: an application configuration '
        'class (preferred), or a package containing an application.'
    )
)
setting_home_view = namespace.add_setting(
    default=DEFAULT_COMMON_HOME_VIEW, global_name='COMMON_HOME_VIEW',
    help_text=_(
        'Name of the view attached to the brand anchor in the main menu. '
        'This is also the view to which users will be redirected after '
        'log in.'
    )
)
setting_project_title = namespace.add_setting(
    default=DEFAULT_COMMON_PROJECT_TITLE, global_name='COMMON_PROJECT_TITLE',
    help_text=_(
        'Name to be displayed in the main menu.'
    )
)
setting_project_url = namespace.add_setting(
    default=DEFAULT_COMMON_PROJECT_URL, global_name='COMMON_PROJECT_URL',
    help_text=_(
        'URL of the installation or homepage of the project.'
    )
)
setting_url_base_path = namespace.add_setting(
    default=DEFAULT_COMMON_URL_BASE_PATH, global_name='COMMON_URL_BASE_PATH',
    help_text=_('Base URL path to use for all views.')
)



from django.apps import apps

from ..managers import TrashCanManager

from .document_models import Document

__all__ = ('DeletedDocument',)


class DeletedDocument(Document):
    objects = TrashCanManager()

    class Meta:
        proxy = True

    def get_api_image_url(self, *args, **kwargs):
        """
        Override the get_api_image_url to allow trashed documents to still
        provide an URL to their first page for preview while in the trash
        can.
        """
        latest_version = self.latest_version
        if latest_version:
            first_page = self.pages.filter(enabled=True).first()

            if first_page:
                return first_page.get_api_image_url(*args, **kwargs)

    @property
    def pages_valid(self):
        """
        Override the pages_valid to allow trashed documents to still provide
        an accurate count of their pages while in the trash can.
        """
        try:
            return self.latest_version.pages.filter(enabled=True)
        except AttributeError:
            # Document has no version yet
            DocumentPage = apps.get_model(
                app_label='documents', model_name='DocumentPage'
            )

            return DocumentPage.objects.none()



import os

from django.conf import settings

DEFAULT_ERROR_LOG_PARTITION_ENTRY_LIMIT = 3
DEFAULT_LOGGING_ENABLE = True
DEFAULT_LOGGING_HANDLERS = ('console',)
DEFAULT_LOGGING_LEVEL = 'ERROR'
DEFAULT_LOGGING_LOG_FILE_PATH = os.path.join(settings.MEDIA_ROOT, 'error.log')

LOGGING_HANDLER_OPTIONS = ('console', 'logfile')
